-- MODÈLE

create schema $model
;;


create table $model.configuration(
    id serial primary key,
    name varchar(24) unique not null default project.unique_name('CFG_'),
    description varchar
)
;;

insert into $model.configuration(name) values('default') 
;;

create table $model.metadata(
    id integer not null default 1 unique check (id=1),
    version varchar(24) not null default '$version',
    creation_date timestamp not null default current_date,
    configuration integer not null references $model.configuration(id)
)
;;

insert into $model.metadata(version, configuration) 
select pm.version, 1 from hydra.metadata as pm;
;;

create table $model.generation_step(
    id serial primary key,
    name varchar(24) unique not null default project.unique_name('GEN_STEP_'),
    description varchar
)
;;

-- TABLES SERVANT A LA CONSTRUCTION

-- CONTAINEURS

create table $model._node(
    id serial primary key,
    node_type hydra_node_type not null,
       unique (id, node_type),
    name varchar(24) unique not null,
    geom geometry('POINTZ',$srid) unique not null check(ST_IsValid(geom)),
    generated integer references $model.generation_step(id) on delete cascade,
    configuration json,
    validity boolean
)
;;

create index ${model}_node_geom_idx on $model._node using gist(geom)
;;
create index ${model}_node_name_idx on $model._node(name)
;;
create index ${model}_node_generated_idx on $model._node(generated)
;;

-- LIEN

create table $model._link(
    id serial primary key,
    link_type hydra_link_type not null,
       unique (id, link_type),
    name varchar(24) unique not null,
    generated integer references $model.generation_step(id) on delete cascade,
    up integer,
    up_type hydra_node_type,
        foreign key (up, up_type) references $model._node(id, node_type) on delete set null,
    down integer,
    down_type hydra_node_type,
        foreign key (down, down_type) references $model._node(id, node_type) on delete set null,
        check(down_type != $catchment_node_type),
        check(up is not null or up_type = $elem_2d_node_type),
        check(down is not null or down_type = $elem_2d_node_type),
    geom geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom)),
    configuration json,
    validity boolean
)
;;

create index ${model}_link_geom_idx on $model._link using gist(geom)
;;

create index ${model}_link_name_idx on $model._link(name)
;;

create index ${model}_link_generated_idx on $model._link(generated)
;;

create index ${model}_link_up_idx on $model._link(up)
;;

create index ${model}_link_down_idx on $model._link(down)
;;


/* common function to get endpoints from geometry
raise error if endpoints cannot be found*/
create function ${model}.find_link_fct(    
    in geom_ geometry('LINESTRINGZ',$srid),
    out up_ integer,
    out up_type_ hydra_node_type,
    out down_ integer,
    out down_type_ hydra_node_type)
language plpgsql
as $$$$
    begin
        select id, node_type from $model._node
        where ST_DWithin(ST_StartPoint(geom_), geom, .1) 
        limit 1 into up_, up_type_
        ;
        select id, node_type from $model._node
        where ST_DWithin(ST_EndPoint(geom_), geom, .1) 
        limit 1 into down_, down_type_
        ;
        if (up_ is null) then
            -- find link from contours
            with node as (
                select id, 'urban_storage' as node_type, contour from $model._urban_storage_node
                union
                select id, 'storage' as node_type, contour from $model._storage_node
                union
                select id, 'elem_2d' as node_type, contour from $model._elem_2d_node
            )
            select id, node_type from node
            where ST_Intersects(ST_StartPoint(geom_), contour) 
            limit 1 into up_, up_type_
            ;
        end if;
        if (down_ is null) then
            -- find link from contours
            with node as (
                select id, 'urban_storage' as node_type, contour from $model._urban_storage_node
                union
                select id, 'storage' as node_type, contour from $model._storage_node
                union
                select id, 'elem_2d' as node_type, contour from $model._elem_2d_node
            )
            select id, node_type from node
            where ST_Intersects(ST_EndPoint(geom_), contour) 
            limit 1 into down_, down_type_
            ;
        end if;

        if (up_ is null or down_ is null) then
            raise exception 'link % not touching nodes (distance from start % distance from end %)', ST_AsText(geom_), (select ST_Distance(ST_StartPoint(geom_), geom) from $model._node order by ST_Distance(ST_StartPoint(geom_), geom) limit 1), (select ST_Distance(ST_EndPoint(geom_), geom) from $model._node order by ST_Distance(ST_EndPoint(geom_), geom) limit 1);
        end if;
    end;
$$$$
;;

-- SINGULARITIES

/* this abstraction is here to do the type checks and to garanty unicity 
of singularity on a given container through the unicity of 'id' */
create table $model._singularity(
    id serial primary key,
    singularity_type hydra_singularity_type not null,
        unique(id, singularity_type),
    name varchar(24) unique not null,
    node integer unique not null,
    node_type hydra_node_type not null,
        foreign key(node, node_type) references $model._node(id, node_type),
    configuration json,
    validity boolean
)
;;

create index ${model}_singularity_name_idx on $model._singularity(name)
;;

/* update geometry of inter_paves sylbols after update */
create function ${model}.inter_pave_symbol_fct(
    in up integer,
    in down integer,
    out out_ geometry('LINESTRINGZ',$srid)
)
language plpgsql
as $$$$
    declare
       pstart_ geometry('POINTZ',$srid);
       pend_ geometry('POINTZ',$srid);
       tx_ real;
       ty_ real;
    begin
       select geom from $model._node where id=up into pstart_;
       select geom from $model._node where id=down into pend_;
       select .1*(ST_X(pend_)-ST_X(pstart_)), .1*(ST_Y(pend_)-ST_Y(pstart_)) into tx_, ty_;
       with p1 as (select (ST_DumpPoints(contour)).geom::geometry as geom from $model._elem_2d_node where id=up),
       p2 as (select (ST_DumpPoints(contour)).geom::geometry as geom from $model._elem_2d_node where id=down),
       p1_ as (select distinct geom from p1),
       p2_ as (select distinct geom from p2),
       pairs as (select ST_Collect(p1_.geom, p2_.geom) as geom from p1_, p2_ order by ST_Distance(p1_.geom, p2_.geom) limit 2),
       line as (select ST_Collect(geom) as geom from pairs),
       symb as (select ST_Centroid(geom) as geom from line)       
       select ST_SetSRID(ST_MakeLine(ST_MakePoint(ST_X(geom)-tx_, ST_Y(geom)-ty_, ST_Z(pstart_)), ST_MakePoint(ST_X(geom)+tx_, ST_Y(geom)+ty_, ST_Z(pend_))), $srid) from symb into out_;
    end;
$$$$
;;

create function ${model}.inter_pave_border_fct(
    in up integer,
    in down integer,
    out out_ geometry('LINESTRINGZ',$srid)
)
language plpgsql
as $$$$
    begin
        with border as (
            select ST_CollectionHomogenize(
                ST_Intersection(n1.contour, n2.contour)) as geom, 0.5*(ST_Z(n1.geom) + ST_Z(n2.geom)) as z 
            from ${model}.elem_2d_node as n1, ${model}.elem_2d_node as n2
            where n1.id=up and n2.id=down
        )
        select ST_SetSRID(ST_MakeLine(
                ST_MakePoint(ST_X(ST_StartPoint(geom)), ST_Y(ST_StartPoint(geom)), z),
                ST_MakePoint(ST_X(ST_EndPoint(geom)), ST_Y(ST_EndPoint(geom)), z)), $srid)
        from border into out_;
    end;
$$$$
;;

/* add trigger metadata configuration  update to update all configured elements*/
create function ${model}.metadata_configuration_after_update_fct()
returns trigger
language plpgsql
as $$$$
    declare
        node record;
        update_fields varchar;
        config_name varchar;
    begin
        select name from $model.configuration where id=new.configuration into config_name;
        for node in select * from $model._node where configuration is not null loop
            if config_name not in (select k from json_object_keys(node.configuration) as k) then
                select 'default' into config_name;
            end if;
            select string_agg(column_name||'=(configuration_json->'''||config_name||'''->>'''||column_name||''')::'||data_type, ', ')
            from information_schema.columns 
            where table_schema='$model' 
            and table_name='_'||node.node_type::varchar||'_node' 
            and data_type != 'USER-DEFINED' 
            and column_name not in ('id', 'node_type')
            into update_fields
            ;
            execute 'update $model.'||node.node_type::varchar||'_node '||
                         'set '||update_fields||' where id='||node.id::varchar||';'
            ;
        end loop;

        return new;
    end;
$$$$
;;

create trigger ${model}_metadata_configuration_after_update_trig
    before update of configuration on $model.metadata
       for each row execute procedure ${model}.metadata_configuration_after_update_fct()
;;

create function ${model}.add_configuration_fct(in configuration json, in id integer, in table_name varchar) returns void
language plpgsql
as $$$$
    declare
        config varchar;
    begin
        if configuration is not null then
            -- @todo complete configuration with default if needed
            if 'default' not in (select k from json_object_keys(configuration) as k) then
                raise exception 'trying to add a configuration with no default values';
            end if;
            insert into $model.configuration(name) 
                select k 
                from json_object_keys(configuration) as k 
                where k not in (select name from $model.configuration);
        end if;
    end;
$$$$
;;




