create table $model.catchment(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('CATCHMENT_'),
    geom geometry('POLYGON',$srid) not null check(ST_IsValid(geom))
)
;;

create table $model.urban_flood_risk_area(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('UFRA_'),
    geom geometry('POLYGONZ',$srid)
)
;;

create table $model.bank(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('BANK_'),
    geom geometry('LINESTRINGZ',$srid) not null
)
;;

create table $model._constrain(
    id serial primary key,
    name varchar(16) unique not null,
    geom geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom)),
    discretized geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom)),
    elem_length real not null default 100,
    constrain_type hydra_constrain_type,
    link_attributes json
)
;;

create table $model.coverage(
    id serial primary key,
    domain_type hydra_coverage_type,
    geom geometry('POLYGONZ',$srid) not null check(ST_IsValid(geom))
)
;;

create table $model.reach(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('REACH_'),
    pk0_km real not null default 0,
    dx real not null default 50,
    geom geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom))
)
;;

create index ${model}_reach_geom_idx on $model.reach using gist(geom)
;;

create table $model.street(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('STREET_'),
    width real not null default 8,
    elem_length real,
    rk real not null default 40.,
    geom geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom))
)
;;

create index ${model}_street_geom_idx on $model.street using gist(geom)
;;


create table $model.station(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('STATION_'),
    geom geometry('POLYGONZ',$srid) not null check(ST_IsValid(geom))
)
;;

/* add trigger on stations to update nodes validities */
create function ${model}.station_after_fct()
returns trigger
language plpgsql
as $$$$
    begin
        if tg_op = 'INSERT' then
            update $model.river_node
                set geom=geom
                where st_intersects(geom, new.geom);
            update $model.manhole_node
                set geom=geom
                where st_intersects(geom, new.geom);
            return new;
        else
            if tg_op = 'UPDATE' then
                update $model.river_node
                    set geom=geom
                    where st_intersects(geom, old.geom) or st_intersects(geom, new.geom);
                update $model.manhole_node
                    set geom=geom
                    where st_intersects(geom, old.geom) or st_intersects(geom, new.geom);
                return new;
            else
                update $model.river_node
                    set geom=geom
                    where st_intersects(geom, old.geom);
                update $model.manhole_node
                    set geom=geom
                    where st_intersects(geom, old.geom);
                return old;
            end if;
        end if;
    end;
$$$$
;;

create trigger ${model}_station_trig
    after insert or update or delete on $model.station
       for each row execute procedure ${model}.station_after_fct()
;;

create table $model.domain_2d(
    id serial primary key,
    name varchar(16) unique not null default project.unique_name('DOMAIN_2D_'),
    mesh_implement_mode varchar default 'strict' check(mesh_implement_mode in ('loose', 'strict')),
    geom geometry('POLYGONZ',$srid)
)
;;

create index ${model}_domain_2d_geom_idx on $model.domain_2d using gist(geom)
;;

create table $model.branch(
    id integer primary key,
    name varchar(24) unique,
    pk0_km real not null default 0,
    dx real not null default 50,
    geom geometry('LINESTRINGZ',$srid) not null check(ST_IsValid(geom))
)
;;

create function $model.branch_update_fct()
returns boolean
language plpgsql
as $$$$
    begin
        update $model._pipe_link set branch = null;
        delete from $model.branch;

        -- count up and down node connectivity on pipes
        -- select pass_trhough manhole nodes that have one up and one down
        -- linemerge pipes that have pass_through at both ends
        -- add pipes that are touching

        with up as (
            select n.id, count(1) as ct, n.geom
            from $model.pipe_link as l join $model.manhole_node as n on n.id=l.up
            group by n.id, n.geom
        ),
        down as (
            select n.id, count(1) as ct
            from $model.pipe_link as l join $model.manhole_node as n on n.id=l.down
            group by n.id
        ),
        blade as (
            select st_collect(n.geom) as geom
            from $model.manhole_node as n
            where id in (
            select id from $model.manhole_node
            except
            select up.id
            from up join down on up.id=down.id
            where up.ct = 1
            and down.ct = 1)
        ),
        branch as (
            select (st_dump(st_split(st_linemerge(st_collect(l.geom)), b.geom))).geom as geom
            from $model.pipe_link as l, blade as b
            where l.up_type='manhole'
            group by b.geom
        )
        insert into $model.branch(id, name, geom)
        select row_number() over(), 'BRANCH_'||(row_number() over())::varchar, geom
        from branch
        ;

        update $model.branch as b set name=m.name, dx=m.dx, pk0_km=m.pk0_km
        from $model.pipe_branch_marker_singularity as m where st_intersects(m.geom, b.geom);

        update $model._pipe_link as p set branch = b.id
        from $model._link as l, $model.branch as b
        where p.id = l.id and st_covers(b.geom, l.geom);

        -- remove hydrology branch
        return 't';
    end;
$$$$
;;




