-- -- Tables annexes
--

create index ${model}_constrain_geom_idx on $model._constrain using gist(geom)
;;

create index ${model}_constrain_discretized_geom_idx on $model._constrain using gist(discretized)
;;

create view $model.constrain as
select id, name, geom, discretized, elem_length, constrain_type, link_attributes
from $model._constrain
;;

-- the aim of the editable view is to allow grid snapping and auto-discretization to avoid
-- topological problems
-- note that the discretized is not editable
create function ${model}.constrain_fct()
returns trigger
language plpgsql
as $$$$
    declare
        id_ integer;
        prec_ real;
        snapped_ geometry;
        discretized_ geometry;
    begin
        if tg_op = 'INSERT' then
            select precision from hydra.metadata into prec_;
            select st_snaptogrid(new.geom, prec_) into snapped_;
            select st_snaptogrid(project.discretize_line(new.geom, coalesce(new.elem_length, 100)), prec_) into discretized_;
            insert into $model._constrain(name, geom, discretized, elem_length, constrain_type, link_attributes)
            values (coalesce(new.name, 'define_later'), snapped_, discretized_, coalesce(new.elem_length, 100), new.constrain_type, new.link_attributes)
            returning id, elem_length into new.id, new.elem_length;
            update $model._constrain set name = 'CONSTR'||new.id::varchar where name = 'define_later' and id = new.id
            returning name into new.name;
            return new;
        elsif tg_op = 'UPDATE' then
            select precision from hydra.metadata into prec_;
            select st_snaptogrid(new.geom, prec_) into snapped_;
            select st_snaptogrid(project.discretize_line(new.geom, new.elem_length), prec_) into discretized_;
            update $model._constrain set name=new.name, geom=snapped_, discretized=discretized_,
                elem_length=new.elem_length, constrain_type=new.constrain_type, link_attributes=new.link_attributes
                where id = old.id;
            return new;
        elsif tg_op = 'DELETE' then
            delete from $model._constrain where id=old.id;
            return old;
        end if;
    end;
$$$$
;;

create trigger ${model}_constrain_trig
    instead of insert or update or delete on $model.constrain
       for each row execute procedure ${model}.constrain_fct()
;;

create index ${model}_coverage_geom_idx on $model.coverage using gist(geom)
;;

create or replace view $model.close_point as
    with allpoints as (
        select id, (st_dumppoints(geom)).geom as geom from $model.constrain
    ),
    ends as (
        select id, st_startpoint(geom) as geom from $model.constrain
        union
        select id, st_endpoint(geom) as geom from $model.constrain
    ),
    cpoints as (
        select id, st_collect(geom) as geom from allpoints group by id
    ),
    clust as (
        select
          sqrt(ST_Area(ST_MinimumBoundingCircle(gc)) / pi()) as radius,
          ST_NumGeometries(gc) as nb,
          ST_Centroid(gc) as geom
        from (
          select unnest(ST_ClusterWithin(geom, 5)) gc
          from allpoints
        ) f
        union
        select
            1 as radius,
            0 as nb,
            st_force2d(e.geom) as geom
        from $model.constrain as c, cpoints as cp, ends as e
        where st_dwithin(c.geom, e.geom, 1)
        and not st_intersects(cp.geom, e.geom)
        and e.id != c.id
        and c.id = cp.id
    )
select row_number() over () AS id, radius, nb, geom::geometry('POINT', $srid) from clust where radius > 0
;;

-- VALIDITY
create view $model.invalid as
with node_invalidity_reason as ($node_invalidity_reason),
link_invalidity_reason as ($link_invalidity_reason),
singularity_invalidity_reason as ($singularity_invalidity_reason)
select CONCAT('node:', n.id) as id, validity, name, 'node'::varchar as type, node_type::varchar as subtype, geom::geometry('POINTZ', $srid) as geom, reason from $model._node as n join node_invalidity_reason as r on r.id=n.id where validity=FALSE
union
select CONCAT('link:', l.id) as id, validity, name, 'link'::varchar as type, link_type::varchar as subtype, ST_Force3D(st_centroid(geom))::geometry('POINTZ', $srid) as geom, reason from $model._link as l join link_invalidity_reason as r on  r.id=l.id where validity=FALSE
union
select CONCAT('singularity:', s.id) as id, s.validity, s.name, 'singularity'::varchar as type, s.singularity_type::varchar as subtype, n.geom::geometry('POINTZ', $srid) as geom, reason from $model._singularity as s join $model._node as n on s.node = n.id join singularity_invalidity_reason as r on r.id=s.id where s.validity=FALSE
union
select CONCAT('profile:', p.id) as id, p.validity, p.name, 'profile'::varchar as type, ''::varchar as subtype, n.geom::geometry('POINTZ', $srid) as geom, ''::varchar as reason from $model._river_cross_section_profile as p join $model._node as n on p.id = n.id where p.validity=FALSE
;;

/* add rule to create surfaces after edit constrain */
create function ${model}.constrain_after_fct()
returns trigger
language plpgsql
as $$$$
    begin
        -- remove from coverage the surafces that do not belong to domains
        delete from $model.coverage ; -- where domain_2d is null;
        alter sequence $model.coverage_id_seq restart;

        with cst as (
            select
                    (st_dump(
                        coalesce(
                            st_split(
                                a.discretized,
                                (select st_collect(discretized)
                                    from $model.constrain as b
                                    where a.id!=b.id and st_intersects(a.discretized, b.discretized))),
                            a.discretized)
                    )).geom as geom
            from $model.constrain as a

        )
        -- polygonize
        insert into $model.coverage(geom, domain_type)
        select (st_dump(st_polygonize(geom))).geom, '2d' from cst;


        with oned as (
            select c.id, count(1) as ct
            from $model.coverage as c, $model.street as d
            where st_intersects(c.geom, d.geom)
            group by c.id
        )
        update $model.coverage set domain_type='street' where id in (select oned.id from oned where ct=1);

        with oned as (
            select c.id, count(1) as ct
            from $model.coverage as c, $model.street as d
            where st_intersects(c.geom, d.geom)
            group by c.id
        )
        update $model.coverage set domain_type='crossroad' where id in (select oned.id from oned where ct>1);

        update $model.coverage as c set domain_type='reach'
        where exists (select 1 from $model.open_reach as r
        where st_intersects(c.geom, r.geom)
        and st_length(st_collectionextract(st_intersection(c.geom, r.geom), 2)) > 0);

        update $model.coverage as c set domain_type='reach_end'
        where exists (select 1 from $model.open_reach as r
            where (st_intersects(c.geom, st_startpoint(r.geom)) and st_distance(st_startpoint(r.geom), st_exteriorring(c.geom)) > 5)
            or (st_intersects(c.geom,  st_endpoint(r.geom))) and st_distance(st_endpoint(r.geom), st_exteriorring(c.geom)) > 5);

        with oned as (
            select distinct c.id
            from $model.coverage as c, $model.storage_node as d
            where st_intersects(c.geom, d.geom)
        )
        update $model.coverage set domain_type='storage' where id in (select oned.id from oned);
        with oned as (
            select distinct c.id
            from $model.coverage as c, $model.urban_storage_node as d
            where st_intersects(c.geom, d.geom)
        )
        update $model.coverage set domain_type='storage' where id in (select oned.id from oned);

        if tg_op = 'INSERT' or tg_op = 'UPDATE'  then
            return new;
        else
            return old;
        end if;


    end;
$$$$
;;

create trigger ${model}_constrain_after_trig
    after insert or update of geom, elem_length or delete on $model._constrain
       execute procedure ${model}.constrain_after_fct()
;;


/* add rule to create constrains after inser of streer */
create function ${model}.street_after_insert_fct()
returns trigger
language plpgsql
as $$$$
    begin
        alter table $model._constrain disable trigger ${model}_constrain_after_trig;
        with buf as (
            select st_buffer(st_force2d(new.geom), new.width/2, 'quad_segs=1 join=mitre mitre_limit='||(new.width*2)::varchar) as geom
        ),
        cst as (
            select st_exteriorring(geom) as geom from buf
        ),
        snp as (
            update $model.constrain set geom = st_snap(geom, new.geom, 1)
        ),
        spl as (
            select coalesce(
                st_split(cst.geom, (select st_collect(st_force2d(c.geom))
                        from $model.constrain as c
                        where st_intersects(c.geom, cst.geom)
                        )), cst.geom) as geom
            from cst
        ),
        spld as (
            select st_force3d((st_dump(geom)).geom) as geom from spl
        ),
        diff as (
            select id, (st_dump(st_collectionextract(st_split(c.geom, b.geom), 2))).geom as geom
            from $model.constrain as c, cst as b
            where st_intersects(b.geom, c.geom)
        ),
        ins as (
            insert into $model.constrain(geom, elem_length, constrain_type)
            select d.geom, c.elem_length, c.constrain_type
            from $model.constrain as c, diff as d
            where c.id = d.id
            union
            select
            geom,
            coalesce(new.elem_length, 100),
            'strickler'
            from spld
            --where not st_intersects(spld.geom, new.geom)
            returning id
        )
        delete from $model.constrain where id in (select id from diff);

        -- for some reason, the following executed here will remove all interesting constrains
        --delete from $model.constrain as c where (select count(1) from $model.street as s where st_intersects(s.geom, c.geom)) > 0;

        alter table $model._constrain enable trigger ${model}_constrain_after_trig;
        return new;

    end;
$$$$
;;

create trigger ${model}_street_after_insert_trig
    after insert on $model.street
       for each row execute procedure ${model}.street_after_insert_fct()
;;


create function ${model}.mesh(domain_id integer)
returns integer
language plpython3u
as
$$$$
    import plpy
    from $hydra import mesh
    return mesh(plpy, domain_id, '$model', $srid)
$$$$
;;

create function ${model}.create_links(up_domain_id integer, down_domain_id integer)
returns integer
language plpython3u
as
$$$$
    import plpy
    from $hydra import create_links
    return create_links(plpy, up_domain_id, down_domain_id, '$model', $srid)
$$$$
;;

create function ${model}.create_elem2d_links(elem_id integer, border geometry, generation_step integer default null)
returns integer
language plpython3u
as
$$$$
    import plpy
    from $hydra import create_elem2d_links
    return create_elem2d_links(plpy, elem_id, border, '$model', $srid, generation_step)
$$$$
;;

/* add rule to update constrain and coverage after domain_2d modification */
create function ${model}.domain_2d_after_fct()
returns trigger
language plpgsql
as $$$$
    begin
        if tg_op = 'INSERT' or tg_op = 'UPDATE'  then
            update $model.elem_2d_node as n
            set domain_2d=(select id from $model.domain_2d as d
                where st_intersects(n.geom, d.geom) and d.id=new.id);
            return new;
        else
            return old;
        end if;
    end;
$$$$
;;

create trigger ${model}_domain_2d_after_trig
    after insert or update or delete on $model.domain_2d
       for each row execute procedure ${model}.domain_2d_after_fct()
;;

create function ${model}.set_link_altitude(link_id integer)
returns integer
language plpython3u
as
$$$$
    import plpy
    from $hydra import set_link_altitude
    return set_link_altitude(plpy, link_id, '$model', $srid)
$$$$
;;

/* add rule to update street_links when street nodes change */
create function ${model}.update_street_links()
returns integer
language plpgsql
as $$$$
    declare
        prec_ real;
        res_ integer;
    begin
        select precision from hydra.metadata into prec_;

        delete from $model.street_link
        where generated is not null;

        with ins as (
            select rk, width, st_makeline(
                lag(n.geom)
                over (partition by s.id order by st_linelocatepoint(s.geom, n.geom)),
                n.geom) as geom, n.generated
            from $model.street as s, $model.crossroad_node as n
            where st_dwithin(s.geom, n.geom, prec_)
        ),
        gen as (
            insert into $model.generation_step default values returning id
        ),
        links as (
            insert into $model.street_link(rk, width, geom, generated)
            select rk, width, geom, gen.id from ins, gen where geom is not null
            returning id
        )
        select count(1) from links into res_;
        return res_;
    end;
$$$$
;;

-- for any point in reach coverage, give the transect interpolated at this point
-- for points on the border bewteen two reach coverage, you can specify the coverage id
create function ${model}.interpolate_transect_at(point geometry, coverage_id integer default null)
returns geometry('LINESTRINGZ', $srid)
language plpgsql
as $$$$
    declare
        l_ real;
    begin
        l_ = 10000;
        --raise warning 'interpolate_transect_at % %', point, coverage_id;
        return (
            with reach as (
                select r.id, r.geom, c.geom as flood_plain
                from $model.coverage as c, $model.reach as r
                where st_intersects(r.geom, c.geom)
                and domain_type='reach'
                and coalesce(c.id = coverage_id, 't'::boolean)
                order by c.geom <-> point
                limit 1
            ),
            transect_piece as (
                select st_linemerge(st_intersection(c.discretized, r.flood_plain)) as geom
                from $model.constrain as c, reach as r
                where st_intersects(r.flood_plain, c.discretized)
                and st_intersects(r.geom, c.geom)
                and c.constrain_type='flood_plain_transect'
                and st_length(st_collectionextract(st_intersection(c.discretized, r.flood_plain), 2)) > 0
            ),
            transect as (
                select st_makeline(
                    st_translate(st_startpoint(c.geom), -l_*sin(2*pi()-st_azimuth(st_startpoint(c.geom), st_endpoint(c.geom))), l_*cos(2*pi()-st_azimuth(st_startpoint(c.geom), st_endpoint(c.geom)))),
                    st_translate(st_startpoint(c.geom), l_*sin(2*pi()-st_azimuth(st_startpoint(c.geom), st_endpoint(c.geom))), -l_*cos(2*pi()-st_azimuth(st_startpoint(c.geom), st_endpoint(c.geom))))
                ) as geom, st_azimuth(st_startpoint(c.geom), st_endpoint(c.geom)) as azi
                from transect_piece as c
            ),
            dist as (
                select (case when azi > pi() then azi-2*pi() else azi end) as azi, st_distance(point, geom) as d from transect
            ),
            ct as  (
                select count(1) as n from dist
            ),
            disttot as (
                select sum(d) as sumd from dist
            ),
            interp as (
                select case when ct.n=1 then (
                    select azi from dist
                )
                when ct.n=2 then (
                    select sum(azi*(1-d/sumd)) as azi
                    from dist, disttot
                )
                end as azi
                from ct
            )
            select st_intersection(st_makeline(
                st_translate(point, l_*sin(2*pi()-azi), -l_*cos(2*pi()-azi)),
                st_translate(point, -l_*sin(2*pi()-azi), l_*cos(2*pi()-azi))), flood_plain)
            from interp, reach
        );
    end;
$$$$
;;
/* generate pl1d */
create function ${model}.update_river_cross_section_pl1d(coverage_id integer)
returns integer
language plpgsql
as $$$$
    declare
        prec_ real;
        res_ integer;
        reach_id_ integer;
    begin
        select precision from hydra.metadata into prec_;

        select r.id
        from $model.coverage as c, $model.reach as r
        where st_intersects(r.geom, c.geom)
        and c.id=coverage_id
        and st_intersects(r.geom, c.geom)
        limit 1
        into reach_id_;

        delete from $model.river_cross_section_pl1d
        where generated is not null;

        -- create banks
        drop table if exists banks;
        create  table banks
        as
        with res as (
            select (st_dump(st_linemerge((st_collectionextract(st_intersection(cs.discretized, cv.geom), 2))))).geom as geom
            from $model.coverage as cv, $model.constrain as cs, $model.reach as r
            where st_intersects(r.geom, cv.geom)
            and cv.id=coverage_id
            and r.id = reach_id_
            and not st_intersects(r.geom, cs.geom)
        )
        select row_number() over() as id, geom::geometry('LINESTRINGZ', $srid) from res
        ;


        -- river points that form the bedline are the one
        -- which are the closest point of their projection and the banks
        drop table if exists bedline;
        create  table bedline
        as
        with projected as (
            select b.id as bank, st_lineinterpolatepoint(b.geom, st_linelocatepoint(b.geom, r.geom)) as geom
            from (
                select (st_dumppoints(geom)).geom as geom
                from $model.reach where id=reach_id_
                union
                select n.geom
                from $model.river_node as n
                where n.reach=reach_id_
                ) as r, banks as b
            union
            select b.id as bank, (st_dumppoints(b.geom)).geom as geom
            from banks as b
        ),
        closest as (
            select b.bank, st_lineinterpolatepoint(r.geom, st_linelocatepoint(r.geom, b.geom)) as geom
            from $model.reach as r, projected as b
            where r.id=reach_id_
        )
        select row_number() over() as id, c.bank, st_makeline(c.geom order by st_linelocatepoint(b.geom, c.geom)) as geom
        from closest as c, banks as b
        where c.bank=b.id
        group by c.bank
        ;

        drop table if exists midline;
        create  table midline
        as
        select row_number() over() as id, st_makeline(st_centroid(st_makeline(p.geom, st_closestpoint(b.geom, p.geom)))) as geom, p.bank
        from banks as b, (select (st_dumppoints(geom)).geom as geom, bank from bedline) as p
        where b.id=p.bank
        group by p.bank
        ;

        drop table if exists pl1d;
        create  table pl1d
        as
        select  row_number() over() as id,
            st_makeline(st_lineinterpolatepoint((select geom from banks where id = 1), i::real/10),
                        st_lineinterpolatepoint((select geom from banks where id = 2), i::real/10)) as geom,
                    1 as node
        from generate_series(1, 10) as i
        ;
        return 1;
    end;
$$$$
;;

create function ${model}.geom_dat(type_ varchar, id integer)
returns varchar
language plpython3u
as
$$$$
    import plpy
    from $hydra import geom_dat
    return geom_dat(plpy, '$model', type_, id)
$$$$
;;

create function ${model}.ctrz_dat()
returns varchar
language plpython3u
as
$$$$
    import plpy
    from $hydra import ctrz_dat
    return ctrz_dat(plpy, '$model')
$$$$
;;

create view ${model}.open_reach
as
with ends as (
    select p.id, p.geom, 'down' as typ, r.id as reach, st_linelocatepoint(r.geom, p.geom) as alpha
    from ${model}.river_cross_section_profile as p, ${model}.reach as r
    where p.type_cross_section_up = 'valley' and p.type_cross_section_down != 'valley' and st_dwithin(p.geom, r.geom, .1)
    union
    select p.id, p.geom, 'up' as typ, r.id as reach, st_linelocatepoint(r.geom, p.geom) as alpha
    from ${model}.river_cross_section_profile as p, ${model}.reach as r
    where p.type_cross_section_up != 'valley' and p.type_cross_section_down = 'valley' and st_dwithin(p.geom, r.geom, .1)
    union
    select p.id, p.geom, 'down' as typ, r.id as reach, st_linelocatepoint(r.geom, p.geom) as alpha
    from ${model}.river_cross_section_profile as p, ${model}.reach as r
    where p.type_cross_section_up = 'valley' and st_dwithin(p.geom, r.geom, .1) and st_linelocatepoint(r.geom, p.geom) > .99
    union
    select p.id, p.geom, 'up' as typ, r.id as reach, st_linelocatepoint(r.geom, p.geom) as alpha
    from ${model}.river_cross_section_profile as p, ${model}.reach as r
    where p.type_cross_section_down = 'valley' and st_dwithin(p.geom, r.geom, .1) and st_linelocatepoint(r.geom, p.geom) < .01
)
,
bounds as (
    select FIRST_VALUE(reach) over (partition by reach order by alpha), typ, LAG(alpha) over (partition by reach order by alpha) as start, alpha as end, id, reach from ends
)
select row_number() over() as id, st_linesubstring(r.geom, b.start, b.end) as geom, r.id as reach from bounds as b join ${model}.reach as r on r.id=b.reach where b.typ = 'down'
;;

-- plpython3u doesn't support multidim arrays yet, so we pass it as string and cast
-- hence the use of _street_after_insert_trig

create function ${model}._str_profile_from_transect(flood_plain_transect geometry, interpolate boolean default 'f'::boolean)
returns varchar
language plpython3u immutable
as
$$$$
    import plpy
    from $hydra import profile_from_transect
    return profile_from_transect(plpy, flood_plain_transect, '$model', $srid, interpolate)
$$$$
;;

create function ${model}.profile_from_transect(flood_plain_transect geometry, interpolate boolean default 'f'::boolean)
returns real[][]
language plpgsql immutable
as
$$$$
    begin
        return $model._str_profile_from_transect(flood_plain_transect, interpolate)::real[][];
    end;
$$$$
;;

-- closest section on the same reach
create function $model.upstream_valley_section(transect geometry)
returns integer
language plpgsql immutable
as
$$$$
    begin
        return (
        with reach as (
            select r.id, st_intersection(r.geom, transect) as inter, geom, pk0_km
            from ${model}.reach as r
            where st_intersects(r.geom, transect)
        ),
        pk as (
            select st_linelocatepoint(geom, inter)*st_length(geom)/1000+pk0_km as pk_km from reach
        ),
        up_reach as (
            select nup.reach as id
            from ${model}.river_node as nup
            join ${model}.connector_link as l on l.up=nup.id
            join ${model}.river_node as ndown on ndown.id=l.down
            where l.main_branch
            and ndown.reach = (select id from reach)
        ),
        extended_reach_node as (
            select n.id, n.pk_km
            from ${model}.river_node as n, pk
            where n.reach in (select id from reach union select id from up_reach)
            and n.pk_km <= pk.pk_km
        ),
        section_up as (
            select c.id
            from extended_reach_node as n 
            join ${model}.river_cross_section_profile as c on c.id = n.id, 
            pk
            where ((c.type_cross_section_up='valley' and c.up_vcs_geom is not null) 
                or (c.type_cross_section_down='valley' and c.down_vcs_geom is not null))
            order by n.pk_km desc 
            limit 1
        )
        select id from section_up);
    end;
$$$$
;;

create view ${model}.l2d
as
with link as (
    select l.name, project.crossed_border(l.geom, n.contour) as geom 
    from ${model}._link as l, ${model}.elem_2d_node as n 
    where (n.id = l.up or n.id=l.down) 
    and l.link_type!='mesh_2d' 
    and l.link_type!='network_overflow')
select name, st_x(st_startpoint(geom)) as xa, st_y(st_startpoint(geom)) as ya, st_x(st_endpoint(geom)) as xb, st_y(st_endpoint(geom)) as yb from link
;;

create function $model.downstream_valley_section(transect geometry)
returns integer
language plpgsql immutable
as
$$$$
    begin
        return (
        with reach as (
            select r.id, st_intersection(r.geom, transect) as inter, geom, pk0_km
            from ${model}.reach as r
            where st_intersects(r.geom, transect)
        ),
        pk as (
            select st_linelocatepoint(geom, inter)*st_length(geom)/1000+pk0_km as pk_km from reach
        ),
        down_reach as (
            select ndown.reach as id
            from ${model}.river_node as ndown
            join ${model}.connector_link as l on l.down=ndown.id
            join ${model}.river_node as nup on nup.id=l.up
            where l.main_branch
            and nup.reach = (select id from reach)
        ),
        extended_reach_node as (
            select n.id, n.pk_km
            from ${model}.river_node as n, pk
            where n.reach in (select id from reach union select id from down_reach)
            and n.pk_km >= pk.pk_km
        ),
        section_down as (
            select c.id
            from extended_reach_node as n 
            join ${model}.river_cross_section_profile as c on c.id = n.id, 
            pk
            where ((c.type_cross_section_down='valley' and c.down_vcs_geom is not null) 
                or (c.type_cross_section_up='valley' and c.up_vcs_geom is not null))
            order by n.pk_km asc 
            limit 1
        )
        select id from section_down);
    end;
$$$$
;;
