--create server multicorn_hydra_fwd_w15 foreign data wrapper multicorn
--options (
--  wrapper 'hydra_fdw.W15ForeignDataWrapper'
--)
--;;

create table project.model_config(
    id serial primary key,
    name varchar(16) not null,
    config integer not null default 1,
        unique(name, config),
    cascade_order integer not null default currval('project.model_config_id_seq'),
    comment varchar(256)
)
;;

create table project.dem(
    id serial primary key,
    priority integer not null default currval('project.dem_id_seq'),
    source varchar(256) unique not null
)
;;

create function project.altitude(x real, y real)
returns real
language plpython3u immutable
as
$$$$
    from $hydra import altitude
    import plpy

    sources = [res['source'].replace('\\','\\\\') for res in plpy.execute("select source from project.dem order by priority desc")]
    return altitude(plpy, sources, x, y)
$$$$
;;

create function project.altitude(point geometry)
returns real
language plpgsql immutable
as
$$$$
    begin
        return project.altitude(st_x(point)::real, st_y(point)::real);
    end;
$$$$
;;

create function project.crossed_border(line geometry, polygon geometry)
returns geometry
language plpython3u immutable
as
$$$$
    from $hydra import crossed_border
    import plpy
    return crossed_border(line, polygon)
$$$$
;;

create function project.crest_line(line geometry)
returns geometry
language plpython3u immutable
as
$$$$
    from $hydra import crest_line
    import plpy

    sources = [res['source'].replace('\\','\\\\') for res in plpy.execute("select source from project.dem order by priority desc")]
    return crest_line(plpy, sources, line, $srid)
$$$$
;;

create function project._str_filling_curve(polygon geometry)
returns varchar
language plpython3u immutable
as
$$$$
    from $hydra import filling_curve
    import plpy

    sources = [res['source'].replace('\\','\\\\') for res in plpy.execute("select source from project.dem order by priority desc")]
    return filling_curve(sources, polygon)
$$$$
;;

-- plpython functions cannot return multidim arrays
create function project.filling_curve(polygon geometry)
returns real[][]
language plpgsql immutable
as
$$$$
    begin
        return project._str_filling_curve(polygon)::real[][];
    end;
$$$$
;;



create function project.set_altitude(geom geometry)
returns geometry
language plpython3u immutable
as
$$$$
    from $hydra import set_altitude
    import plpy
    sources = [res['source'].replace('\\','\\\\') for res in plpy.execute("select source from project.dem order by priority desc")]
    return set_altitude(plpy, sources, geom, $srid)
$$$$
;;



create table project.grp(
    id serial primary key,
    name varchar(256) unique not null default 'GROUP_'||currval('project.grp_id_seq')::varchar
)
;;

create table project.grp_model(
    grp integer not null references project.grp(id) on delete cascade,
    model integer not null references project.model_config(id),
        unique(grp, model)
)
;;

create table project.output_option(
    id serial primary key,
    sor1 bool not null default 'f',
    sor2 bool not null default 'f',
    sor3 bool not null default 'f',
    sor7 bool not null default 'f',
    sor10 bool not null default 'f',
    nosorw13 bool not null default 'f',
    nosorw14 bool not null default 'f',
    nosorw15 bool not null default 'f',
    nosorw16 bool not null default 'f'
)
;;


create table project.scenario(
    id serial primary key,
    name varchar(24) not null default project.unique_name('SCN_'),
    comment varchar(256),
    comput_mode hydra_computation_mode not null default 'hydrology_and_hydraulics',
    dry_inflow integer references project.dry_inflow_scenario(id),
    rainfall integer references project._rainfall(id),
    dt_hydrol_mn real not null default 5,
    soil_moisture_coef real default 0 check(soil_moisture_coef >= 0 and soil_moisture_coef <=1),
    runoff_adjust_coef real default 1 check(runoff_adjust_coef >= 0),
    option_dim_hydrol_network bool not null default false,
    date0 timestamp not null default '2000-01-01 12:00:00',
    tfin_hr real not null default 12,
    tini_hydrau_hr real not null default 12,
    dtmin_hydrau_hr real not null default 0.005,
    dtmax_hydrau_hr real not null default 0.1,
    dzmin_hydrau real not null default 0.05,
    dzmax_hydrau real not null default 0.1,
    flag_save bool not null default 'f',
    tsave_hr real default 0,
    flag_rstart bool not null default 'f',
    flag_gfx_control bool not null default 'f',
    scenario_rstart integer references project.scenario(id),
    flag_hydrology_rstart bool not null default 'f',
    scenario_hydrology_rstart integer references project.scenario(id),
    trstart_hr real default 0,
    graphic_control bool not null default 'f',
    model_connect_settings hydra_model_connection_settings not null default 'cascade',
    tbegin_output_hr real not null default 0,
    tend_output_hr real not null default 999.e12,
    dt_output_hr real not null default .1,
    c_affin_param bool not null default 'f',
    t_affin_start_hr real check(c_affin_param = 'f' or t_affin_start_hr is not null),
    t_affin_min_surf_ddomains real not null default 0,
    strickler_param bool not null default 'f',
    zini_param bool not null default 'f',
    option_file bool not null default 'f',
    option_file_path varchar(256),
    output_option integer unique not null references project.output_option(id),
    transport_type hydra_transport_type not null default 'no_transport',
    iflag_mes bool not null default 'f',
    iflag_dbo5 bool not null default 'f',
    iflag_dco bool not null default 'f',
    iflag_ntk bool not null default 'f',
    quality_type hydra_quality_type not null default 'physico_chemical',
    diffusion bool not null default 'f',
    dispersion_coef real not null default 20,
    longit_diffusion_coef real not null default 41,
    lateral_diff_coef real not null default .45,
    water_temperature_c real not null default 20,
    min_o2_mgl real not null default 1.5,
    kdbo5_j_minus_1 real not null default 0.2,
    koxygen_j_minus_1 real not null default 0.45,
    knr_j_minus_1 real not null default 0.09,
    kn_j_minus_1 real not null default 0.035,
    bdf_mgl real not null default 1.5,
    o2sat_mgl real not null default 9.07,
    iflag_ad1 bool not null default 'f',
    iflag_ad2 bool not null default 'f',
    iflag_ad3 bool not null default 'f',
    iflag_ad4 bool not null default 'f',
    ad1_decay_rate_jminus_one real not null default 0,
    ad2_decay_rate_jminus_one real not null default 0,
    ad3_decay_rate_jminus_one real not null default 0,
    ad4_decay_rate_jminus_one real not null default 0,
    sst_mgl real not null default 0,
    sediment_file varchar(256) not null default ''
)
;;

create table project.mixed(
    ord integer not null,
    scenario integer not null references project.scenario(id) on delete cascade,
    grp integer not null references project.grp(id),
        unique(scenario, grp),
        unique(scenario, ord)
)
;;

create table project.param_hydrograph(
    id serial primary key,
    param_scenario integer not null references project.scenario(id) on delete cascade,
    model varchar(24) not null,
    hydrograph_from varchar(24) not null,
    hydrograph_to varchar(256) not null
)
;;

create table project.param_external_hydrograph(
    id serial primary key,
    param_scenario integer not null references project.scenario(id) on delete cascade,
    external_file varchar(256) not null
)
;;

create table project.param_regulation(
    id serial primary key,
    param_scenario integer not null references project.scenario(id) on delete cascade,
    control_file varchar(256) not null
)
;;

create table project.param_measure(
    id serial primary key,
    param_scenario integer not null references project.scenario(id) on delete cascade,
    measure_file varchar(256) not null
)
;;

create table project.inflow_rerouting(
    id serial primary key,
    param_scenario integer not null references project.scenario(id) on delete cascade,
    model varchar(24) not null,
    hydrograph_from varchar(24) not null,
    hydrograph_to varchar(256) not null,
    option hydra_rerouting_option not null,
    value real not null
)
;;

create table project.zini_param(
    id serial primary key,
    scenario integer not null references project.scenario(id) on delete cascade,
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null,
    zini real not null
)
;;


create table project.c_affin_param(
    id serial primary key,
    scenario integer not null references project.scenario(id) on delete cascade,
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null,
    smin real not null
)
;;

create table project.strickler_param(
    id serial primary key ,
    scenario integer not null references project.scenario(id) on delete cascade,
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null,
    rkmin real not null,
    rkmaj real not null,
    pk1 real not null,
    pk2 real not null
)
;;

create table project.sor1(
    id serial primary key ,
    poo integer not null references project.output_option(id),
    model integer not null references project.model_config(id),
    name_obj varchar(24) not null,
    type_sor1 hydra_type_sor1 not null,
    param_type hydra_param_sor1_type not null,
    name_us varchar(24) not null
)
;;

create table project.sor2(
    id serial primary key ,
    poo integer not null references project.output_option(id),
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null
)
;;

create table project.sor3(
    id serial primary key ,
    poo integer not null references project.output_option(id),
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null
)
;;

create table project.sor7(
    id serial primary key ,
    poo integer not null references project.output_option(id),
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null
)
;;

create table project.sor10(
    id serial primary key ,
    poo integer not null references project.output_option(id),
    model varchar(24) not null,
    type_domain varchar(24) not null check(type_domain in ('reach', 'domain_2d', 'branch')),
    element varchar(24) not null
)
;;



create table project.scenario_group(
    id serial primary key,
    name varchar(24) unique not null default 'GROUP_'||currval('project.scenario_group_id_seq')::varchar
)
;;

create table project.param_scenario_group(
    scenario integer not null references project.scenario(id),
    grp integer not null references project.scenario_group(id),
        unique(scenario, grp)
)
;;

create table project.hydrograph_variable_dry_flow(
    id serial primary key,
    sector integer not null references project.dry_inflow_sector(id),
    distrib_coef real not null,
    lag_time_hr real not null default 0,
    dry_inflow_hourly_modulation integer not null references project.dry_inflow_hourly_modulation(id)
)
;;


create table project.hydrograph_pollution(
    id serial primary key,
    dry_flow_mes_mgl real not null,
    dry_flow_dbo5_mgl real not null,
    dry_flow_dco_mgl real not null,
    dry_flow_ntk_mgl real not null,
    runoff_mes_mgl real not null,
    runoff_flow_dbo5_mgl real not null,
    runoff_flow_dco_mgl real not null,
    runoff_flow_ntk_mgl real not null
)
;;

create table project.hydrograph_quality(
    id serial primary key,
    dry_flow_dbo5_mgl real not null,
    dry_flow_nh4_mgl real not null,
    dry_flow_no3_mgl real not null,
    dry_flow_o2_mgl real not null,
    dry_flow_ad_mgl real not null,
    dry_flow_mes_mgl real not null,
    runoff_flow_dbo5_mgl real not null,
    runoff_flow_nh4_mgl real not null,
    runoff_flow_no3_mgl real not null,
    runoff_flow_ad_mgl real not null,
    runoff_flow_mes_mgl real not null,
    dbou_dbo5_ratio real not null default 1.2,
    dbo5_dcant_fraction real not null default 0,
    mes_decant_fraction real not null default 0
)
;;


create table project.bc_tz_data(
    id serial primary key,
    filename varchar(256) not null
)
;;

create or replace function hydra_to_json(tabl varchar, geom varchar, variadic cols varchar[])
returns varchar
language plpgsql
as $$$$
    declare
        res varchar;
        ct integer;
    begin
        execute 'select count(1) from '||tabl into ct;

        if ct>0 and geom is null then
            execute 'select row_to_json(everything)
            from
                (select ''FeatureCollection'' as type,
                        array_to_json(array_agg(feat)) as features
                 from (select ''Feature'' as type,
                           null As geometry,
                           row_to_json((select l from (select '||(select array_to_string(cols, ','))||') as l)) as properties
                       from '||tabl||'
                      ) as feat
                 ) as everything;' into res;
        elsif ct>0 then
            execute 'select row_to_json(everything)
            from
                (select ''FeatureCollection'' as type,
                        (select row_to_json(l) from (select ''name'' as type,
                            (select row_to_json(m)
                                from (select ''urn:ogc:def:crs:EPSG::''||(select ST_SRID('||geom||') from '||tabl ||' limit 1)::varchar as name) as m)
                            as properties) as l)
                        as crs,
                        array_to_json(array_agg(feat)) as features
                 from (select ''Feature'' as type,
                           ST_AsGeoJSON('||geom||')::json As geometry,
                           row_to_json((select l from (select '||(select array_to_string(cols, ','))||') as l)) as properties
                       from '||tabl||'
                      ) as feat
                 ) as everything;' into res;
        else
            select '' into res;
        end if;
        return res;
    end;
$$$$
;;

-- Does not work with 3D GEOM
create function project.discretize_line(geom geometry, elem_length float)
returns geometry
language plpython3u
as $$$$
    import plpy
    from $hydra import discretize_line
    return discretize_line(plpy, geom, elem_length)
$$$$
;;

create function project.create_result(scenario varchar, model varchar)
returns bool
language plpython3u
as $$$$
    import plpy
    from $hydra import create_result
    workspace = plpy.execute("select workspace from hydra.metadata")[0]["workspace"]
    for statement in create_result(workspace, scenario, model, $srid):
        plpy.execute(statement)
    return True
$$$$
;;



