# coding = utf-8

from subprocess import Popen, PIPE
from io import StringIO
import os
from shapely import wkb
from shapely.geometry import Point
import numpy


topo = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'topo.exe' if os.name != 'posix' else 'topo')

def zb(s_left_bank, s_right_bank, topo_profile, pk=None, pk_up=None, pk_down=None, zb_up=None, zb_down=None):
    """returns [zb_left]+[zb_min]+[zb_right] array (list of 4+6+4 2-tuple)
    for the given topographic profile
    s_right_bank and s_right_bank are location of the bank along the profile
    given in in km
    pk, pk_up, pk_down are the location of current, up and down cross-sections
    zb_up, zb_down are the zb arrays for the up and down cross-sections
    """

    in_ = "{s_left_bank} {s_right_bank} {nbpt}\n{topo_profile}\n\n{mode}\n".format(
            s_left_bank=s_left_bank,
            s_right_bank=s_right_bank,
            nbpt=len(topo_profile),
            topo_profile="\n".join(["{} {}".format(p[0], p[1]) for p in topo_profile]),
            mode=0 if pk is None else 1
            )
    if pk:
        in_ += "{pk_up} {pk} {pk_down}\n{nbptl}\n{ptl}\n{nbptr}\n{ptr}\n\n".format(
                pk_up=pk_up,
                pk=pk,
                pk_down=pk_down,
                nbptl=len(zb_up),
                ptl="\n".join(["{} {}".format(p[0], p[1]) for p in zb_up]),
                nbptr=len(zb_down),
                ptr="\n".join(["{} {}".format(p[0], p[1]) for p in zb_down])
                )
    out, err = Popen([topo], stdout=PIPE, stdin=PIPE, stderr=PIPE).communicate(bytes(in_, 'utf-8'))
    if err:
        raise Exception(err)
    lines = out.split(b'\n')
    zb_min = [(float(z), float(b)) for z, b in zip(lines[0].split(), lines[1].split())]
    zb_majl = [(float(z), float(b)) for z, b in zip(lines[2].split(), lines[3].split())]
    zb_majr = [(float(z), float(b)) for z, b in zip(lines[4].split(), lines[5].split())]
    return zb_majl+zb_min+zb_majr

def pg_array_to_list(arr):
    return [[float(v) for v in coord.split(',')] for coord in arr.replace('{{','').replace('}}','').split('},{')]

def profile_from_transect(plpy, transect, model, srid, interpolate=False):
    """returns a string representation of the array,
    the riverbed is taken fron the two transect points closest to the reach intersection
    a single point of intersection is assumed"""

    #@todo find upstream and downstream profiles in this query
    inter = plpy.execute("""
        with reach as (
            select r.id, st_intersection(r.geom, '{transect}'::geometry) as inter, geom, pk0_km
            from {model}.reach as r
            where st_intersects(r.geom, '{transect}'::geometry)
        ),
        pk as (
            select st_linelocatepoint(geom, inter)*st_length(geom)/1000+pk0_km as pk from reach
        ),
        topo_transect as (
            select project.set_altitude(
                project.discretize_line('{transect}'::geometry, st_length('{transect}'::geometry)/150)
                ) as geom
        ),
        up_ as (
            select coalesce((select zbmin_array from {model}.valley_cross_section_geometry as g where g.id = p.down_vcs_geom),
                            (select zbmin_array from {model}.valley_cross_section_geometry as g where g.id = p.up_vcs_geom)) as zbmin,
                   n.pk_km
            from {model}.river_cross_section_profile as p
            join {model}.river_node as n on p.id=n.id
            where p.id = {model}.upstream_valley_section('{transect}'::geometry)
        ),
        down_ as (
            select coalesce((select zbmin_array from {model}.valley_cross_section_geometry as g where g.id = p.up_vcs_geom),
                            (select zbmin_array from {model}.valley_cross_section_geometry as g where g.id = p.down_vcs_geom)) as zbmin,
                   n.pk_km
            from {model}.river_cross_section_profile as p
            join {model}.river_node as n on p.id=n.id
            where p.id = {model}.downstream_valley_section('{transect}'::geometry)
        )
        select r.id as reach, pk.pk, r.inter, t.geom as topo_transect, (select zbmin from up_)::varchar as su,  (select pk_km from up_) as pku,  (select zbmin from down_)::varchar as sd, (select pk_km from down_) as pkd
        from reach as r, topo_transect as t, pk
        """.format(
            model=model,
            transect=transect
            ))
    if not inter:
        return None
    inter_reach = wkb.loads(inter[0]['inter'], True)
    transect_geom = wkb.loads(transect, True)
    s_center = transect_geom.project(inter_reach)
    s = numpy.array([transect_geom.project(Point(p)) for p in transect_geom.coords])
    s_left_bank, s_right_bank = numpy.sort(s[numpy.argsort(numpy.abs(s - s_center))[:2]])

    topo_profile = [(transect_geom.project(Point(p)), p[2]) for p in wkb.loads(inter[0]['topo_transect'], True).coords]

    #plpy.warning("intersection at ", s_center, s, s_left_bank, s_right_bank, inter[0]['su'], inter[0]['sd'])

    if interpolate and inter[0]['su'] is not None and inter[0]['sd'] is not None:
        pk, pk_up, pk_down =  inter[0]['pk'], inter[0]['pku'], inter[0]['pkd']
        zb_up, zb_down = pg_array_to_list(inter[0]['su']), pg_array_to_list(inter[0]['sd'])
        #plpy.warning("intepolate", pk, pk_up, pk_down, zb_up, zb_down)
        return str(zb(s_left_bank, s_right_bank, topo_profile,
            pk, pk_up, pk_down, zb_up, zb_down)).replace('[', '{').replace('(', '{').replace(']', '}').replace(')', '}')
    else:
        return str(zb(s_left_bank, s_right_bank, topo_profile)).replace('[', '{').replace('(', '{').replace(']', '}').replace(')', '}')